#define STRICT
#define DIRECTINPUT_VERSION 0x0800

#include <windows.h>
#include <d3dx9.h>
#include <dinput.h>

#include "system.h"

//[of]:input
//[of]:InputManager
#define KEYDOWN(name, key) (name[key] & 0x80)

extern HINSTANCE	hInstance	;
extern HWND	mainWindow	;

//[of]:constructor
InputManager::InputManager()
{
	directInput	= NULL;
	mouseDevice	= NULL;
	keyboardDevice	= NULL;
}
//[cf]
//[of]:destructor
InputManager::~InputManager()
{
	if( keyboardDevice ) 
	{
		keyboardDevice->Unacquire();
		keyboardDevice->Release();
	}
	
	if( mouseDevice )
	{
		mouseDevice->Unacquire();
		mouseDevice->Release();
	}
	
	if( directInput )
		directInput->Release();
	
}
//[cf]
//[of]:initialize
bool InputManager::initialize()
{
	HRESULT hr = DirectInput8Create(	hInstance,
		DIRECTINPUT_VERSION, 
		IID_IDirectInput8, 
		(void**)&directInput, 
		NULL); 
	if FAILED(hr) 
		return false;
	
	// Acquire keyboard
	hr = directInput->CreateDevice(GUID_SysKeyboard, &keyboardDevice, NULL); 
	if FAILED(hr) 
		return false;
  
	hr = keyboardDevice->SetDataFormat(&c_dfDIKeyboard); 
	if FAILED(hr)
		return false;
  
	hr = keyboardDevice->SetCooperativeLevel(mainWindow, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE); 
	if FAILED(hr)
		return false;
  
	// Acquire mouse
	hr = directInput->CreateDevice(GUID_SysMouse, &mouseDevice, NULL); 
	if FAILED(hr) 
		return false;
	
	hr = mouseDevice->SetDataFormat(&c_dfDIMouse2); 
	if FAILED(hr)
		return false;
	
	hr = mouseDevice->SetCooperativeLevel(mainWindow, DISCL_FOREGROUND | DISCL_EXCLUSIVE); 
	if FAILED(hr)
		return false;
  
	currentBuffer	= keyBuffer1;
	lastBuffer	= keyBuffer2;
	for(int i=0; i<256; i++)
	{
		keyBuffer1[i] = 0;
		keyBuffer2[i] = 0;
	}
	
	return true;
}
//[cf]

//[of]:update
void InputManager::update()
{
//[c]	
//[c]	Keyboard
//[c]	
	if( currentBuffer == keyBuffer1 )
	{
		currentBuffer	= keyBuffer2;
		lastBuffer	= keyBuffer1;
	}
	else
	{
		currentBuffer	= keyBuffer1;
		lastBuffer	= keyBuffer2;
	}
	
	HRESULT hr = mouseDevice->Acquire(); 
	if FAILED(hr)
		return;
	
	HRESULT hr2 = keyboardDevice->Acquire(); 
	hr = keyboardDevice->GetDeviceState(256, (LPVOID)currentBuffer); 
	if FAILED(hr)
		return;
	
//[c]	
//[c]	 Mouse
//[c]	 
	DIMOUSESTATE2 dims2;
	ZeroMemory(&dims2, sizeof(dims2));
	hr = mouseDevice->GetDeviceState(sizeof(DIMOUSESTATE2), &dims2);
	if FAILED(hr)
		return;
	
	mouseState.lastLmb	= mouseState.lmb;
	mouseState.dx	= dims2.lX;
	mouseState.dy	= dims2.lY;
	mouseState.lmb	= (dims2.rgbButtons[0] & 0x80) != 0;
	mouseState.rmb	= (dims2.rgbButtons[1] & 0x80) != 0;
	mouseState.mmb	= (dims2.rgbButtons[2] & 0x80) != 0;
}
//[cf]

//[of]:isKeyPressed
int InputManager::isKeyPressed(int key)
{
	return KEYDOWN(currentBuffer, key) && !KEYDOWN(lastBuffer, key);
}
//[cf]
//[of]:isKeyDown
int InputManager::isKeyDown(int key)
{
	return KEYDOWN(currentBuffer, key);
}
//[cf]
//[of]:getMouseState
MouseState* InputManager::getMouseState()
{
	return &mouseState;
}
//[cf]
//[cf]
//[of]:MouseState
//[of]:constructor
MouseState::MouseState()
{
	dx	= 0;
	dy	= 0;
	lmb	= false;
	mmb	= false;
	rmb	= false;
	lastLmb	= false;
}
//[cf]
//[of]:leftClick
bool MouseState::leftClick()
{
	return lmb && !lastLmb;
}
//[cf]
//[cf]
//[cf]
//[of]:graphics
//[of]:VertexBuffer
//[of]:instance creation
//[of]:constructor
VertexBuffer::VertexBuffer()
{
	handle	= NULL;
}
//[cf]
//[of]:destructor
VertexBuffer::~VertexBuffer()
{
	release();
}
//[cf]
//[cf]
//[of]:creating
//[of]:create
bool VertexBuffer::create(Device* device, dword size, dword format, bool dynamic)
{
	dword usage = D3DUSAGE_WRITEONLY;
	if( dynamic )
		usage |= D3DUSAGE_DYNAMIC;
	
	return !FAILED( device->CreateVertexBuffer(	size,
		usage, 
		format,
		D3DPOOL_MANAGED, 
		&handle,
		NULL ) );
}
//[cf]
//[of]:release
void VertexBuffer::release()
{
	if( handle )
	{
		handle->Release();
		handle = NULL;
	}
}
//[cf]
//[cf]
//[of]:locking
//[of]:lock
void* VertexBuffer::lock()
{
	void* buffer;
	handle->Lock(0, 0, &buffer, 0);
	return buffer;
}
//[cf]
//[of]:unlock
void VertexBuffer::unlock()
{
	handle->Unlock();
}

//[cf]
//[cf]
//[cf]
//[of]:IndexBuffer
//[of]:instance creation
//[of]:constructor
IndexBuffer::IndexBuffer()
{
	handle	= NULL;
}
//[cf]
//[of]:destructor
IndexBuffer::~IndexBuffer()
{
	release();
}
//[cf]
//[cf]
//[of]:creating
//[of]:create
bool IndexBuffer::create(Device* device, dword size)
{
	return !FAILED( device->CreateIndexBuffer(	size,
		D3DUSAGE_WRITEONLY, 
		D3DFMT_INDEX16,
		D3DPOOL_MANAGED, 
		&handle,
		NULL ) );
}
//[cf]
//[of]:release
void IndexBuffer::release()
{
	if( handle )
	{
		handle->Release();
		handle = NULL;
	}
}
//[cf]
//[cf]
//[of]:locking
//[of]:lock
word* IndexBuffer::lock()
{
	void* buffer;
	handle->Lock(0, 0, &buffer, 0);
	return (word*)buffer;
}
//[cf]
//[of]:unlock
void IndexBuffer::unlock()
{
	handle->Unlock();
}
//[cf]
//[cf]
//[cf]
//[of]:Light
//[of]:instance creation
//[of]:constructor
Light::Light()
{
	ZeroMemory(&_light, sizeof(_light));
	
	// Default to directional
	_light.Type	= D3DLIGHT_DIRECTIONAL;
	
	// Default color to white
	setDiffuse(1.0f, 1.0f, 1.0f);
}
//[cf]
//[cf]
//[of]:accessing
//[of]:setDiffuse
void Light::setDiffuse(float r, float g, float b)
{
	_light.Diffuse.r	= r;
	_light.Diffuse.g	= g;
	_light.Diffuse.b	= b;
}
//[cf]
//[of]:setDirection
void Light::setDirection(const Vector& direction)
{
	Vector n = direction;
	n.normalize();
	_light.Direction.x	= n.x;
	_light.Direction.y	= n.y;
	_light.Direction.z	= n.z;
}
//[cf]
//[cf]
//[cf]
//[of]:Material
//[of]:instance creation
//[of]:constructor
Material::Material()
{
	ZeroMemory(&_material, sizeof(_material));
	
	_material.Diffuse.r	= 1.0f;
	_material.Diffuse.g	= 1.0f;
	_material.Diffuse.b	= 1.0f;
	_material.Diffuse.a	= 1.0f;
		
	_material.Ambient.r	= 1.0f;
	_material.Ambient.g	= 1.0f;
	_material.Ambient.b	= 1.0f;
	_material.Ambient.a	= 1.0f;
}
//[cf]
//[cf]
//[cf]
//[of]:Texture
//[of]:instance creation
//[of]:constructor
Texture::Texture()
{
	handle	= NULL;
	width	= 0;
	height	= 0;
}
//[cf]
//[of]:destructor
Texture::~Texture()
{
	if( handle )
		handle->Release();
}
//[cf]
//[cf]
//[of]:loading
//[of]:load
bool Texture::load(Device* device, const char* filename, int _width, int _height, int depth)
{
	width	= _width;
	height	= _height;
	
	if FAILED(D3DXCreateTextureFromFileEx(	device, 
		filename,
		width, 
		height, 
		1, 
		0, 
		(D3DFORMAT)depth, 
		D3DPOOL_MANAGED, 
		D3DX_DEFAULT, 
		D3DX_DEFAULT, 
		0, 
		NULL, 
		NULL, 
		&handle))
	{
		return false;
	}
	
	return true;
}
//[cf]
//[cf]
//[cf]
//[of]:GraphicsDevice
HINSTANCE	hInstance	;
HWND	mainWindow	;

//[of]:instance creation
//[of]:constructor
GraphicsDevice::GraphicsDevice()
{
	mainWindow	= NULL;
	direct3D	= NULL;
	direct3DDevice	= NULL;
	fps	= 0;
}
//[cf]
//[of]:destructor
GraphicsDevice::~GraphicsDevice()
{
	releaseD3D();
	releaseWindow();
}
//[cf]
//[cf]
//[of]:running
//[of]:initialize
bool GraphicsDevice::initialize(const char* _title, int width, int height, bool full)
{
	title	= _title;
	screenWidth	= width;
	screenHeight	= height;
	fullScreen	= full;
	
	return
		initializeWindow(hInstance)	&&
		initializeD3D();
}
//[cf]
//[of]:main
void GraphicsDevice::main()
{
	restore();
	
	MSG msg;
	ZeroMemory(&msg, sizeof(msg));
	DWORD	t1	= GetTickCount();
	int	frames	= 0;
	
	while( true )
	{
		if( PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE) )
		{
			if( msg.message == WM_QUIT )
				return;
	
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else if( ready() )
		{
			frames++;
			DWORD t2 = GetTickCount();
			int d = t2 - t1;
			if( d > 1000 )
			{
				fps = frames;
				t1 = t2;
				frames = 0;
			}
	
			move();
			render();
		}
	}
	
}

//[of]:ready
//[c]Test if the device is ready and reset it if required
//[c]
bool GraphicsDevice::ready()
{
	HRESULT hr = direct3DDevice->TestCooperativeLevel();
	if FAILED(hr)
	{
		// If the device was lost, do not render until we get it back
		if( hr == D3DERR_DEVICELOST )
			return false;
	
		// Check if the device needs to be resized.
		if( hr == D3DERR_DEVICENOTRESET )
		{
			invalidate();
	
			if( FAILED(direct3DDevice->Reset(&d3dpp)) )
				return false;
			if( !restore() )
				return false;
		}
	}
	
	return true;
}
//[cf]
//[cf]
//[of]:quit
void GraphicsDevice::quit()
{
	PostQuitMessage(0);
}
//[cf]
//[cf]
//[of]:accessing
//[of]:device
Device* GraphicsDevice::device() 
{ 
	return direct3DDevice;
}
//[cf]
//[cf]
//[of]:private
//[of]:initializeWindow
//[c]Initialize the main window
//[c]- register the class
//[c]- create the window
//[c]

// Forward delcaration:
LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

bool GraphicsDevice::initializeWindow(HINSTANCE hInstance)
{
	WNDCLASSEX wc = 
	{
		sizeof(WNDCLASSEX), 
		CS_CLASSDC,
		MsgProc, 
		0L, 
		0L, 
		hInstance, 
		NULL, //LoadIcon(hInstance, (LPCSTR)101), 
		NULL, 
		NULL, 
		NULL,
		"GraphicsDevice",
		NULL 
	};
	RegisterClassEx(&wc);
  
	mainWindow =
		CreateWindow( 
			"GraphicsDevice",
			title,
			WS_OVERLAPPEDWINDOW,
			100, 100, screenWidth, screenHeight,
			GetDesktopWindow(), 
			NULL, 
			hInstance, 
			NULL);
  
	ShowWindow(mainWindow, SW_SHOWDEFAULT);
	UpdateWindow(mainWindow);
	
	return true;
}
//[cf]
//[of]:releaseWindow
void GraphicsDevice::releaseWindow()
{
	UnregisterClass("Xracer", hInstance);
}
//[cf]
//[of]:MsgProc
//[c]The window's message handler
//[c]
LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch( msg )
	{
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
		
		case WM_SETCURSOR:
			SetCursor(NULL);
			return TRUE;
		
		case WM_NCHITTEST:
			// Prevent the user from selecting the menu in fullscreen mode
			return HTCLIENT;
		
		case WM_SYSCOMMAND:
			// Prevent moving/sizing and power loss in fullscreen mode
			switch( wParam )
			{
				case SC_MOVE:
				case SC_SIZE:
				case SC_MAXIMIZE:
				case SC_KEYMENU:
				case SC_MONITORPOWER:
				return 1;
			}
			break;
	}
	
	return DefWindowProc(hWnd, msg, wParam, lParam);
}
//[cf]

//[of]:initializeD3D
//[c]Initializes Direct3D
//[c]
//[c]The function returns true if the initialization succeeds.
//[c]
bool GraphicsDevice::initializeD3D()
{
	direct3D = Direct3DCreate9(D3D_SDK_VERSION);
	if( direct3D == NULL )
		return false;
  
	ZeroMemory(&d3dpp, sizeof(d3dpp));
	if( fullScreen )
	{
		D3DFORMAT format = D3DFMT_X8R8G8B8;
		if( !checkMode(format) )
			format = D3DFMT_R5G6B5;
		if( !checkMode(format) )
			return false;
	
		d3dpp.Windowed	= FALSE;
		d3dpp.BackBufferFormat	= format;
		d3dpp.FullScreen_RefreshRateInHz	= D3DPRESENT_RATE_DEFAULT;
	}
	else
	{
		d3dpp.Windowed	= TRUE;
		d3dpp.BackBufferFormat	= D3DFMT_UNKNOWN;
	}
  
	d3dpp.PresentationInterval	= D3DPRESENT_INTERVAL_ONE;
	d3dpp.BackBufferWidth	= screenWidth;
	d3dpp.BackBufferHeight	= screenHeight;
	d3dpp.SwapEffect	= D3DSWAPEFFECT_FLIP;
	d3dpp.BackBufferCount	= 2;
	d3dpp.EnableAutoDepthStencil	= TRUE;
	d3dpp.AutoDepthStencilFormat	= D3DFMT_D16;
	d3dpp.hDeviceWindow	= mainWindow;
	d3dpp.Flags	= D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
  
	D3DCAPS9 d3dCaps;
	if FAILED( direct3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &d3dCaps ) )
		return false;

	DWORD dwBehaviorFlags =
		( d3dCaps.VertexProcessingCaps != 0 ) ?
			D3DCREATE_HARDWARE_VERTEXPROCESSING :
			D3DCREATE_SOFTWARE_VERTEXPROCESSING;

	if( FAILED( direct3D->CreateDevice(	D3DADAPTER_DEFAULT, 
		D3DDEVTYPE_HAL, 
		mainWindow, 
		dwBehaviorFlags, 
		&d3dpp, 
		&direct3DDevice) ) )
	
		return false;
  
	return true;
}

//[c]Sub-functions:
//[of]:	checkMode
//[c]Test if a mode is available
//[c]
bool GraphicsDevice::checkMode(D3DFORMAT format)
{
	DWORD n = direct3D->GetAdapterModeCount(D3DADAPTER_DEFAULT, format);
	for(UINT i=0; i<n; i++)
	{
		D3DDISPLAYMODE mode;
		direct3D->EnumAdapterModes(D3DADAPTER_DEFAULT, format, i, &mode);
		if(	mode.Width == screenWidth &&
			mode.Height == screenHeight )
			return true;
	}
	
	return false;
}
//[cf]
//[cf]
//[of]:releaseD3D
//[c]Releases all previously initialized objects
//[c]
void GraphicsDevice::releaseD3D()
{
	if( direct3DDevice ) 
		direct3DDevice->Release();
	
	if( direct3D )
		direct3D->Release();
}
//[cf]
//[cf]
//[cf]
//[cf]
//[of]:entry point
//[of]:WinMain
//[c]The application's entry point
//[c]
extern int run();

INT WINAPI WinMain(HINSTANCE _hInstance, HINSTANCE, LPSTR, INT )
{
	hInstance = _hInstance;
	return run();
}
//[cf]
//[cf]
