
#pragma warning (disable: 4996)

typedef unsigned char	byte	;
typedef unsigned int	dword	;
typedef unsigned short 	word	;
typedef unsigned int	Color	;
typedef const char*	pcsz	;

#ifndef max
#define max(a,b)	(((a) > (b)) ? (a) : (b))
#define min(a,b)	(((a) < (b)) ? (a) : (b))
#endif

#define NULL 0

//[of]:geometry
//[of]:Math
const float pi	= 3.14159265f;
const float pi2	= 2.0f * pi;
//[cf]
//[of]:Quaternion
class Quaternion : public D3DXQUATERNION
{
public:
	void invert()
	{
		D3DXQuaternionInverse(this, this);
	}
	
};
//[cf]
//[of]:Vector
class Vector
{
public:
	float	x;
	float	y;
	float	z;

public:
	Vector() {}
	
	Vector(const Vector& v) 
	{
		x = v.x;
		y = v.y;
		z = v.z;
	}

	Vector(float _x, float _y, float _z) 
	{
		x = _x;
		y = _y;
		z = _z;
	}
	
	void set(const Vector& v) 
	{
		x = v.x;
		y = v.y;
		z = v.z;
	}
	
	Vector(const Vector& v, float scale) 
	{
		x = scale * v.x;
		y = scale * v.y;
		z = scale * v.z;
	}
	
	void set(float _x, float _y, float _z) 
	{
		x = _x;
		y = _y;
		z = _z;
	}
	
	void set(const Vector& v, float scale) 
	{
		x = scale * v.x;
		y = scale * v.y;
		z = scale * v.z;
	}

	void reset() 
	{
		x = 0.0f;
		y = 0.0f;
		z = 0.0f;
	}
	
	// cross product
	void cp(const Vector& v1, const Vector& v2)
	{
		float tx = v1.y * v2.z - v1.z * v2.y;
		float ty = v1.z * v2.x - v1.x * v2.z;
		float tz = v1.x * v2.y - v1.y * v2.x;
		x = tx;
		y = ty;
		z = tz;
	}
	
	void diff(const Vector& v1, Vector& v2)
	{
		x = v1.x - v2.x;
		y = v1.y - v2.y;
		z = v1.z - v2.z;
	}
	
	void add(const Vector& v, float scalar)
	{
		x += v.x * scalar;
		y += v.y * scalar;
		z += v.z * scalar;
	}

	void scale(float scalar)
	{
		x *= scalar;
		y *= scalar;
		z *= scalar;
	}
	
	float operator* (Vector& v) const
	{
		return x*v.x + y*v.y + z*v.z;
	}
	
	void operator+= (const Vector& v)
	{
		x += v.x;
		y += v.y;
		z += v.z;
	}
	
	void translate(const Vector& v)
	{
		x += v.x;
		y += v.y;
		z += v.z;
	}
	
	void operator-= (const Vector& v)
	{
		x -= v.x;
		y -= v.y;
		z -= v.z;
	}
	
	void normalize()
	{
		float r = x * x + y * y + z * z;
		if( r == 0.0f )
		{
			x = 1.0f;
			y = 1.0f;
			z = 1.0f;
		}
		else
		{
			r = sqrtf(r);
			x /= r;
			y /= r;
			z /= r;
		}
	}

	float norm() const
	{
		return sqrtf(x * x + y * y + z * z);
	}
	
	float squaredNorm() const
	{
		return x * x + y * y + z * z;
	}

	void rotate(const Quaternion& q)
	{
		float a   = q.w;
		float b   = q.x;
		float c   = q.y;
		float d   = q.z;
		
		float t2  =  a*b;
		float t3  =  a*c;
		float t4  =  a*d;
		float t5  = -b*b;
		float t6  =  b*c;
		float t7  =  b*d;
		float t8  = -c*c;
		float t9  =  c*d;
		float t10 = -d*d;
		float x2 = 2.0f * ( (t8 + t10)*x + (t6 -  t4)*y + (t3 + t7)*z ) + x;
		float y2 = 2.0f * ( (t4 +  t6)*x + (t5 + t10)*y + (t9 - t2)*z ) + y;
		float z2 = 2.0f * ( (t7 -  t3)*x + (t2 +  t9)*y + (t5 + t8)*z ) + z;
		x = x2;
		y = y2;
		z = z2;
	}

	void setUp(const Quaternion& q)
	{
		float a = q.w;
		float b = q.x;
		float c = q.y;
		float d = q.z;
		
		x =	2.0f * (b * c - a * d);
		y = 1.0f -	2.0f * (b * b + d * d);
		z =	2.0f * (a * b + c * d);
	}
	
	void setDirection(const Quaternion& q)
	{
		float a = q.w;
		float b = q.x;
		float c = q.y;
		float d = q.z;
		
		x = 1.0f -	2.0f * (c * c + d * d);
		y =	2.0f * (a * d + b * c);
		z =	2.0f * (b * d - a * c);
	}

	void setLeft(const Quaternion& q)
	{
		float a   = q.w;
		float b   = q.x;
		float c   = q.y;
		float d   = q.z;
		
		x =	2.0f * ( a*c + b*d );
		y =	2.0f * ( c*d - a*b );
		z = 1.0f -	2.0f * ( b*b + c*c );
	}

	bool isNul()
	{
		return x == 0.0f && y == 0.0f && z == 0.0f;
	}
};
//[cf]
//[of]:Rotation
class Rotation
{
public:
	float	x;
	float	y;
	float	z;
};
//[cf]
//[cf]
//[of]:input
//[of]:MouseState
class MouseState
{
public:
	int	dx	;
	int	dy	;
	bool	lmb	;
	bool	mmb	;
	bool	rmb	;
	bool	lastLmb	;

	MouseState();
	
	bool leftClick();
};
//[cf]
//[of]:InputManager
class InputManager
{
private:
	LPDIRECTINPUT8	directInput	;
	LPDIRECTINPUTDEVICE8	keyboardDevice	;
	LPDIRECTINPUTDEVICE8	mouseDevice	;
			
	char*	currentBuffer	;
	char*	lastBuffer	;
	char	keyBuffer1[256]	;
	char	keyBuffer2[256]	;
	MouseState	mouseState	;

	// input
	bool	initializeDirectInput	();
	void	releaseDirectInput	();

public:
	InputManager();
	~InputManager();
	bool initialize();

	// updating
	void	update();
	
	// accessing
	int	isKeyPressed	(int key);
	int	isKeyDown	(int key);
	MouseState*	getMouseState	();
};
//[cf]
//[cf]
//[of]:graphics
//[of]:Device
typedef IDirect3DDevice9 Device;
//[cf]
//[of]:Light
class Light
{
private:
	D3DLIGHT9 _light;
	
public:
	Light();
	
	void	setDiffuse	(float r, float g, float b);
	void	setDirection	(const Vector& direction);
	
	D3DLIGHT9* handle() { return &_light; }
};
//[cf]
//[of]:Material
class Material
{
private:
	D3DMATERIAL9 _material;
	
public:
	Material();
	
	D3DMATERIAL9* handle() { return &_material; }
};
//[cf]
//[of]:VertexBuffer
class VertexBuffer
{
public:
	LPDIRECT3DVERTEXBUFFER9	handle;
	
	VertexBuffer();
	~VertexBuffer();
	
	bool	create	(Device* device, dword size, dword format, bool dynamic);
	void	release	();
	void*	lock	();
	void	unlock	();
};
//[cf]
//[of]:IndexBuffer
class IndexBuffer
{
public:
	LPDIRECT3DINDEXBUFFER9	handle;
	
	IndexBuffer();
	~IndexBuffer();
	
	bool	create	(Device* device, dword size);
	void	release	();
	word*	lock	();
	void	unlock	();
};
//[cf]
//[of]:Texture
class Texture
{
public:
	int	width	;
	int	height	;
	LPDIRECT3DTEXTURE9	handle	;

public:
	Texture();
	~Texture();
	
	bool load(Device* device, const char* filename, int width, int height, int depth);
};
//[cf]
//[of]:GraphicsDevice
class GraphicsDevice
{
private:
	const char*	title	;
	bool	fullScreen	;
	int	screenWidth	;
	int	screenHeight	;
	LPDIRECT3D9	direct3D	;
	Device*	direct3DDevice	;
	D3DPRESENT_PARAMETERS	d3dpp	;
	
	// main window
	bool initializeWindow(HINSTANCE hInstance);
	void releaseWindow();
	
	// d3d
	bool	initializeD3D	();
	void	releaseD3D	();
	bool	checkMode	(D3DFORMAT format);
	
	// main
	bool ready();
	
protected:
	virtual bool	restore	() = 0;
	virtual void	invalidate	() = 0;
	virtual void	move	() = 0;
	virtual void	render	() = 0;

public:
	int	fps;
	
	GraphicsDevice();
	~GraphicsDevice();
	
	bool	initialize	(const char* title, int width, int height, bool fullScreen);
	void	main	();
	void	quit	();
	
	// accessing
	Device* device();
};
//[cf]
//[cf]
